/*
 * Real time audio munger thingy
 * Copyright (C) 2003
 * Jason Downer
 * comtaco@cableone.net
 *
 * The fft routines at the end are borrowed / modified from the 
 * fxt library.
 * 
 * Because this code includes GPL'd source (the fast fourier 
 * transform routines) this program also falls under the 
 * General Public License, so without furthur ado...
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *                       
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <math.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>

#define RATE   44100 // sampling rate, adjust lower for slower machine

#define MONO   1
#define STEREO 2

#define MAXLDN 13 // largest ldn to try

int dsp_fd;   // file desc for audio
int file_fd;

int size;     // data size
int ldn;      // log2 of data size: ldn=10 means that there are 1024 elements
int ldnm1;    // ldn minus 1, because we have to demux the right and left channels
int step;     // stride of data reads
int halfstep; // half the stride of data reads
float nscale; // for normalization of coefficients 
float *lerp;  // linear interpolation table

double latency;

short *audio_in_buf;
short *audio_out_buf;

float *buf1; // scratch space
float *buf2; 

float *realL;  // fft data, left
float *imagL;
float *realR;  // fft data, right
float *imagR;

//low volume mark for our fft gate
float cutoff=0.00000001;

timer_t timer;
struct itimerspec value1;
struct itimerspec value2;

void fft(float *fr, float *fi, int ldn, int is);


//Neat stuff happens here
void filter(float *real, float *imag) 
{
int i;
float d,e;

memset(imag, 0, sizeof(float)*step);

//transform into frequency spectrum
fft(real,imag,ldnm1,+1);

real[0]=0;
imag[0]=0;
real[1]=0;

for (i=1;i<step;i++)
	{
	real[i] *= nscale;
	imag[i] *= nscale;
	//get power squared from this freq
	d = (real[i]*real[i] + imag[i]*imag[i]);
	//check against threshold
	if (d<cutoff)
		{
		if (cutoff!=0) 
			{
			e=d/cutoff;
			}
		//This ramps it down parabolically
		real[i] *= e;
		imag[i] *= e;
		}
	//This one sucks
	/*if (d<cutoff)
		{
		real[i] = 0;
		imag[i] = 0;
		}
	else
		{
		if (real[i]>0) real[i] -= cutoff; else real[i] += cutoff;
		if (imag[i]>0) imag[i] -= cutoff; else imag[i] += cutoff;
		}
	*/
	}

//transform back
fft(real,imag,ldnm1,-1);
}

/*-- Boring from here on out  --*/

void inittimer()
	{
	if (timer_create(CLOCK_REALTIME, NULL, &timer) < 0)
		{
		printf("Failed to create timer\n");
		exit(1);
		}
	value1.it_interval.tv_sec = 0;
	value1.it_interval.tv_nsec = 0;
	value1.it_value.tv_sec = 1000000;
	value1.it_value.tv_nsec = 0;
	timer_settime(timer,0,&value1,NULL);
	}

void starttimer()
	{
	timer_gettime(timer,&value2);
	}

double stoptimer()
	{
	double diff;
	timer_gettime(timer,&value1);
	diff = value2.it_value.tv_sec - value1.it_value.tv_sec +
		(value2.it_value.tv_nsec - value1.it_value.tv_nsec)/1E9;
	return diff;
	}
	
void open_audio()
  {
  int tmp, i;

	/* Open an output file with raw data in it */
	file_fd = open("out", O_RDWR|O_CREAT|O_TRUNC,0644);
	
	/* Open the audio device
	 * Your mileage may vary. Wildly.*/
  dsp_fd = open("/dev/dsp", O_RDWR);
  if(dsp_fd < 0)
    {
    printf("Opening /dev/dsp failed\n");
    exit(1);
    }

  //fcntl(dsp_fd, F_SETFL, fcntl(dsp_fd, F_GETFL, 0) & ~O_NDELAY);

  i = tmp = AFMT_S16_LE;
  if(ioctl(dsp_fd, SNDCTL_DSP_SAMPLESIZE, &tmp) < 0 || tmp != i)
    {
    printf("Doesn't support 16-bit sample width\n");
	  exit(1);
    }

  i = tmp = 1;
  if((ioctl(dsp_fd, SNDCTL_DSP_STEREO, &tmp) < 0) || tmp != i)
    {
    printf("Doesn't support stereo samples\n");
    exit(1);
    }
        
  /* Set the sample rate */
  tmp = RATE;
  if(ioctl(dsp_fd, SNDCTL_DSP_SPEED, &tmp) < 0)
    {
    printf("Doesn't support desired sample rate\n");
    exit(1);
    }
    
  if (ioctl(dsp_fd, SNDCTL_DSP_SETDUPLEX, 0) == -1) 
	  {
	  printf("Cannot set bidirectional audio");
		exit(1);
		}
  
	tmp = MAXLDN + 1;
	i = tmp;
    
  if(ioctl(dsp_fd, SNDCTL_DSP_SETFRAGMENT, &tmp) < 0)
    {
    printf("Doesn't support %d-byte buffer fragments (%d)\n", 1<<i, i);
    }

	}

void close_audio()
  {
  if(dsp_fd == -1)
    return;
      
	close(dsp_fd);
  close(file_fd);
  dsp_fd = -1;
  }

void setdatablocks(int length)
	{
	int i;
	if (length<2)
		{
		printf("Bad ldn value\n");
		exit(1);
		}

	ldn = length;
	ldnm1 = ldn-1;
	
	size  = 1 << ldn;
	step  = 1 << ldnm1;
	halfstep  = step >> 1;
	
	nscale = 1.0 / size;
	
	//maximum time filter can take
	latency = (double)step / (double)(RATE * STEREO);
	//printf("Latency:%f\n",latency);
	
	realL = (float*)calloc(step, sizeof(float));
	imagL = (float*)calloc(step, sizeof(float));
	realR = (float*)calloc(step, sizeof(float));
	imagR = (float*)calloc(step, sizeof(float));
		
	}

void unsetdatablocks()
	{
	free(realL);
	free(imagL);
	free(realR);
	free(imagR);
	}

void setaudioblocks()
	{
	int i;
	audio_in_buf  = (short*)calloc(step, sizeof(short));
	audio_out_buf = (short*)calloc(step, sizeof(short));
	buf1 = (float*)calloc(step, sizeof(float));
	buf2 = (float*)calloc(size, sizeof(float));

	lerp = (float*)calloc(halfstep, sizeof(float));
	
	for (i=0;i<halfstep;i++)
		{
		float a;
		//Ok, this is no longer a linear lerp for fading, 
		//it's a cubic spline I found in povray's noise functions
		
		//lerp[i]=(float)i/(float)halfstep;
		a=(float)i/(float)(halfstep-1);
		lerp[i]=((a)*(a)*(3.0-(2.0*(a))));
		}
		
	}

void unsetaudioblocks()
	{
	free(audio_in_buf);
	free(audio_out_buf);
	free(buf1);
	free(buf2);
	free(lerp);
	}

int main(int argc, char ** argv)
	{
	int i,j;
	int bytesread;
	int found=0;
	float diff;
	float floatscale=1.0/(1<<15);
	
	printf("Audiomungerthingy. Ctrl-C to stop. Raw audio is put in the file \"out\"\n\n");
	
	inittimer();
	
	for (i=MAXLDN;i>1;i--)
		{
		setdatablocks(i);
		
		starttimer();
		
		//dummy run
		filter(realL,imagL);
		filter(realR,imagR);
		
		diff=stoptimer();
		//printf("Tried block size %d with time %f seconds\n",size,diff);
		
		if (diff<latency)
			{
			printf("Found good approx blocksize for this machine:\n\t%d elements, time %f	< %f seconds\n",size,diff,latency);
			found=1;
			break;
			}
		unsetdatablocks();
		}
	
	if (!found)
		{
		printf("Unable to find small enough blocksize\n");
		exit(1);
		}
		
	open_audio();
	
	setaudioblocks();

	while (1)//forever and ever and ever and ever
		{
		//This is commented out, because we block our reads (wait for data to show up) so that we're not measuring that waiting
		//starttimer();

		bytesread = read(dsp_fd, audio_in_buf, step*sizeof(short));
		if (bytesread < 1) 
			{
			printf("Error, no data available\n");
			break;
			}
		
		//move the tail of float buf2 to its head
		//so we can overlap our ffts
		memcpy(buf2,&buf2[step],step*sizeof(float));
		
		//convert audio to floats
		for (i=0;i<step;i++)
			{
			buf2[step + i]=((float)audio_in_buf[i])*floatscale;
			}

		//copy the float audio data into the fft buffer, demux it too
		//memcpy(real,buf2,sizeof(float)*size);
		for (i=0,j=0;i<size;i+=2,j++)
			{
			realL[j]=buf2[i];
			realR[j]=buf2[i+1];
			}
		
		//do the fft
		filter(realL,imagL);
		filter(realR,imagR);

		//blend the head of this fft with the tail from the last, remux, copy
		for (i=0,j=0;i<step;i+=2,j++)
			{
			audio_out_buf[i]=  (short)( (lerp[j]*realL[j] + lerp[halfstep-j-1]*buf1[i])   * (1<<15) );
			audio_out_buf[i+1]=(short)( (lerp[j]*realR[j] + lerp[halfstep-j-1]*buf1[i+1]) * (1<<15) );
			//copy the tail of this fft to the previous holding array
			buf1[i]  = realL[j+halfstep];
			buf1[i+1]= realR[j+halfstep];
			}

		write(dsp_fd, audio_out_buf,step*sizeof(short));
		write(file_fd, audio_out_buf,step*sizeof(short));
		
		//memcpy(buf1,&real[step],sizeof(float)*step);
		//diff=stoptimer();
		//printf("\t\ttime: %f seconds\n",diff);
		}

	setaudioblocks();
	close_audio();
	unsetdatablocks();
	}
	
/*This code, with minor alterations, is from the GPL'd FXT-2000 Library*/
#define sumdiff(a,b,s,d)    { s=a+b; d=a-b; }
#define sumdiff_05(a,b,s,d) { s=0.5*(a+b); d=0.5*(a-b); }
#define sumdiff2(s,d)       { float t; t=s-d; s+=d; d=t; }
#define cmult(c,s,u,v)      { float t; t=u*s+v*c; u*=c; u-=v*s; v=t; }
#define SWAP(x,y)           { float tmp=x; x=y; y=tmp; }

void revbin_permute(float *fr, long n);

void fft(float *fr, float *fi, int ldn, int is)
/*
 * radix 2 fft a la cooley tukey
 *
 */
	{
  int    n2,ldm,m,mh,j,r;
  int    t1, t2;
  float pi,phi,c,s;
  float ur,vr, ui,vi;

  n2=1<<ldn;
  pi=is*M_PI;

  revbin_permute(fr,n2);
  revbin_permute(fi,n2);

  for (ldm=1; ldm<=ldn; ++ldm)
    {
    m=(1<<ldm);            /* m=2^ldm */
		mh=(m>>1);             /* mh=m/2 */

    phi=pi/(float)(mh);

    for (j=0; j<mh; ++j)
      {
	    float w=phi*(float)j;

	    c=cos(w);
	    s=sin(w);

  	  for (r=0; r<n2; r+=m)
        {
        t1=r+j;
        t2=t1+mh;

        vr=fr[t2]*c-fi[t2]*s;
        vi=fr[t2]*s+fi[t2]*c;

        ur=fr[t1];
        fr[t1]+=vr;
        fr[t2]=ur-vr;

        ui=fi[t1];
        fi[t1]+=vi;
        fi[t2]=ui-vi;
        }
      }
    }
	} 

void revbin_permute(float *fr, long n)
	{
  long m,j;
  for (m=1,j=0; m<n-1; m++)
    {
		long k;
		for(k=n>>1; (!((j^=k)&k)); k>>=1);

		if (j>m)  SWAP(fr[m],fr[j]);
    }
	}
